@@ -1,8 +1,8 @@
 /*
    +----------------------------------------------------------------------+
-   | PHP version 4.0                                                      |
+   | PHP Version 4                                                        |
    +----------------------------------------------------------------------+
-   | Copyright (c) 1997-2001 The PHP Group                                |
+   | Copyright (c) 1997-2003 The PHP Group                                |
    +----------------------------------------------------------------------+
    | This source file is subject to version 2.02 of the PHP license,      |
    | that is bundled with this package in the file LICENSE, and is        |
@@ -12,11 +12,11 @@
    | obtain it through the world-wide-web, please send a note to          |
    | license@php.net so we can mail you a copy immediately.               |
    +----------------------------------------------------------------------+
-   | Authors: Stig Sther Bakken <ssb@guardian.no>                        |
+   | Author: Stig Sther Bakken <ssb@fast.no>                             |
    +----------------------------------------------------------------------+
  */
 
-/* $Id: formatted_print.c,v 1.1.1.3 2001/07/19 00:20:13 zarzycki Exp $ */
+/* $Id: formatted_print.c,v 1.59.2.14 2004/11/15 13:40:31 derick Exp $ */
 
 #include <math.h>				/* modf() */
 #include "php.h"
@@ -25,6 +25,10 @@
 #include "zend_execute.h"
 #include <stdio.h>
 
+#ifdef HAVE_LOCALE_H
+#include <locale.h>
+#endif
+
 #define ALIGN_LEFT 0
 #define ALIGN_RIGHT 1
 #define ADJ_WIDTH 1
@@ -61,6 +65,7 @@ static char HEXCHARS[] = "0123456789ABCDEF";
 static char *php_convert_to_decimal(double arg, int ndigits, int *decpt, int *sign, int eflag)
 {
 	register int r2;
+	int mvl;
 	double fi, fj;
 	register char *p, *p1;
 	/*THREADX*/
@@ -86,13 +91,21 @@ static char *php_convert_to_decimal(double arg, int ndigits, int *decpt, int *si
 		p1 = &cvt_buf[NDIG];
 		while (fi != 0) {
 			fj = modf(fi / 10, &fi);
+			if (p1 <= &cvt_buf[0]) {
+				mvl = NDIG - ndigits;
+				memmove(&cvt_buf[mvl], &cvt_buf[0], NDIG-mvl-1);
+				p1 += mvl;
+			}
 			*--p1 = (int) ((fj + .03) * 10) + '0';
 			r2++;
 		}
 		while (p1 < &cvt_buf[NDIG])
 			*p++ = *p1++;
 	} else if (arg > 0) {
-		while ((fj = arg * 10.0) < 0.9999999) {
+		while ((fj = arg * 10) < 1) {
+			if (!eflag && (r2 * -1) < ndigits) {
+				break;
+			}
 			arg = fj;
 			r2--;
 		}
@@ -105,10 +118,17 @@ static char *php_convert_to_decimal(double arg, int ndigits, int *decpt, int *si
 		cvt_buf[0] = '\0';
 		return (cvt_buf);
 	}
+	if (p <= p1 && p < &cvt_buf[NDIG]) {
+		arg = modf(arg * 10, &fj);
+		if ((int)fj==10) {
+			*p++ = '1';
+			fj = 0;
+			*decpt = ++r2;
+		}
 	while (p <= p1 && p < &cvt_buf[NDIG]) {
-		arg *= 10;
-		arg = modf(arg, &fj);
 		*p++ = (int) fj + '0';
+			arg = modf(arg * 10, &fj);
+		}
 	}
 	if (p1 >= &cvt_buf[NDIG]) {
 		cvt_buf[NDIG - 1] = '\0';
@@ -136,11 +156,11 @@ static char *php_convert_to_decimal(double arg, int ndigits, int *decpt, int *si
 
 
 inline static void
-php_sprintf_appendchar(char **buffer, int *pos, int *size, char add)
+php_sprintf_appendchar(char **buffer, int *pos, int *size, char add TSRMLS_DC)
 {
 	if ((*pos + 1) >= *size) {
 		*size <<= 1;
-		PRINTF_DEBUG(("%s: ereallocing buffer to %d bytes\n", get_active_function_name(), *size));
+		PRINTF_DEBUG(("%s(): ereallocing buffer to %d bytes\n", get_active_function_name(TSRMLS_C), *size));
 		*buffer = erealloc(*buffer, *size);
 	}
 	PRINTF_DEBUG(("sprintf: appending '%c', pos=\n", add, *pos));
@@ -151,11 +171,14 @@ php_sprintf_appendchar(char **buffer, int *pos, int *size, char add)
 inline static void
 php_sprintf_appendstring(char **buffer, int *pos, int *size, char *add,
 						   int min_width, int max_width, char padding,
-						   int alignment, int len, int sign, int expprec)
+						   int alignment, int len, int neg, int expprec, int always_sign)
 {
 	register int npad;
+	int req_size;
+	int copy_len;
 
-	npad = min_width - MIN(len, (expprec ? max_width : len));
+	copy_len = (expprec ? MIN(max_width, len) : len);
+	npad = min_width - copy_len;
 
 	if (npad < 0) {
 		npad = 0;
@@ -163,29 +186,30 @@ php_sprintf_appendstring(char **buffer, int *pos, int *size, char *add,
 	
 	PRINTF_DEBUG(("sprintf: appendstring(%x, %d, %d, \"%s\", %d, '%c', %d)\n",
 				  *buffer, *pos, *size, add, min_width, padding, alignment));
-	if ((max_width == 0) && (! expprec)) {
-		max_width = MAX(min_width, len);
-	}
-	if ((*pos + max_width) >= *size) {
-		while ((*pos + max_width) >= *size) {
+
+	req_size = *pos + MAX(min_width, copy_len) + 1;
+
+	if (req_size > *size) {
+		while (req_size > *size) {
 			*size <<= 1;
 		}
 		PRINTF_DEBUG(("sprintf ereallocing buffer to %d bytes\n", *size));
 		*buffer = erealloc(*buffer, *size);
 	}
 	if (alignment == ALIGN_RIGHT) {
-		if (sign && padding=='0') {
-			(*buffer)[(*pos)++] = '-';
+		if ((neg || always_sign) && padding=='0') {
+			(*buffer)[(*pos)++] = (neg) ? '-' : '+';
 			add++;
 			len--;
+			copy_len--;
 		}
 		while (npad-- > 0) {
 			(*buffer)[(*pos)++] = padding;
 		}
 	}
 	PRINTF_DEBUG(("sprintf: appending \"%s\"\n", add));
-	memcpy(&(*buffer)[*pos], add, MIN(max_width, len)+1);
-	*pos += MIN(max_width, len);
+	memcpy(&(*buffer)[*pos], add, copy_len + 1);
+	*pos += copy_len;
 	if (alignment == ALIGN_LEFT) {
 		while (npad--) {
 			(*buffer)[(*pos)++] = padding;
@@ -195,19 +219,21 @@ php_sprintf_appendstring(char **buffer, int *pos, int *size, char *add,
 
 
 inline static void
-php_sprintf_appendint(char **buffer, int *pos, int *size, int number,
-						int width, char padding, int alignment)
+php_sprintf_appendint(char **buffer, int *pos, int *size, long number,
+						int width, char padding, int alignment, 
+						int always_sign)
 {
 	char numbuf[NUM_BUF_SIZE];
-	register unsigned int magn, nmagn, i = NUM_BUF_SIZE - 1, neg = 0;
+	register unsigned long magn, nmagn;
+	register unsigned int i = NUM_BUF_SIZE - 1, neg = 0;
 
 	PRINTF_DEBUG(("sprintf: appendint(%x, %x, %x, %d, %d, '%c', %d)\n",
 				  *buffer, pos, size, number, width, padding, alignment));
 	if (number < 0) {
 		neg = 1;
-		magn = ((unsigned int) -(number + 1)) + 1;
+		magn = ((unsigned long) -(number + 1)) + 1;
 	} else {
-		magn = (unsigned int) number;
+		magn = (unsigned long) number;
 	}
 
 	/* Can't right-pad 0's on integers */
@@ -218,26 +244,30 @@ php_sprintf_appendint(char **buffer, int *pos, int *size, int number,
 	do {
 		nmagn = magn / 10;
 
-		numbuf[--i] = (magn - (nmagn * 10)) + '0';
+		numbuf[--i] = (unsigned char)(magn - (nmagn * 10)) + '0';
 		magn = nmagn;
 	}
 	while (magn > 0 && i > 0);
 	if (neg) {
 		numbuf[--i] = '-';
+	} else if (always_sign) {
+		numbuf[--i] = '+';
 	}
 	PRINTF_DEBUG(("sprintf: appending %d as \"%s\", i=%d\n",
 				  number, &numbuf[i], i));
 	php_sprintf_appendstring(buffer, pos, size, &numbuf[i], width, 0,
 							 padding, alignment, (NUM_BUF_SIZE - 1) - i,
-							 neg, 0);
+							 neg, 0, always_sign);
 }
 
 inline static void
-php_sprintf_appenduint(char **buffer, int *pos, int *size, int number,
-						int width, char padding, int alignment)
+php_sprintf_appenduint(char **buffer, int *pos, int *size,
+					   unsigned long number,
+					   int width, char padding, int alignment)
 {
 	char numbuf[NUM_BUF_SIZE];
-	register unsigned int magn, nmagn, i = NUM_BUF_SIZE - 1;
+	register unsigned long magn, nmagn;
+	register unsigned int i = NUM_BUF_SIZE - 1;
 
 	PRINTF_DEBUG(("sprintf: appenduint(%x, %x, %x, %d, %d, '%c', %d)\n",
 				  *buffer, pos, size, number, width, padding, alignment));
@@ -251,13 +281,13 @@ php_sprintf_appenduint(char **buffer, int *pos, int *size, int number,
 	do {
 		nmagn = magn / 10;
 
-		numbuf[--i] = (magn - (nmagn * 10)) + '0';
+		numbuf[--i] = (unsigned char)(magn - (nmagn * 10)) + '0';
 		magn = nmagn;
-	}
-	while (magn > 0 && i > 0);
+	} while (magn > 0 && i > 0);
+
 	PRINTF_DEBUG(("sprintf: appending %d as \"%s\", i=%d\n", number, &numbuf[i], i));
 	php_sprintf_appendstring(buffer, pos, size, &numbuf[i], width, 0,
-							 padding, alignment, (NUM_BUF_SIZE - 1) - i, 0, 0);
+							 padding, alignment, (NUM_BUF_SIZE - 1) - i, 0, 0, 0);
 }
 
 inline static void
@@ -265,12 +295,23 @@ php_sprintf_appenddouble(char **buffer, int *pos,
 						 int *size, double number,
 						 int width, char padding,
 						 int alignment, int precision,
-						 int adjust, char fmt)
+						 int adjust, char fmt,
+						 int always_sign
+						 TSRMLS_DC)
 {
 	char numbuf[NUM_BUF_SIZE];
 	char *cvt;
 	register int i = 0, j = 0;
-	int sign, decpt;
+	int sign, decpt, cvt_len;
+	char decimal_point = '.';
+#ifdef HAVE_LOCALE_H
+	struct lconv lc;
+	char locale_decimal_point;
+	localeconv_r(&lc);
+	locale_decimal_point = (lc.decimal_point)[0];
+#else
+	char locale_decimal_point = '.';
+#endif
 
 	PRINTF_DEBUG(("sprintf: appenddouble(%x, %x, %x, %f, %d, '%c', %d, %c)\n",
 				  *buffer, pos, size, number, width, padding, alignment, fmt));
@@ -283,43 +324,80 @@ php_sprintf_appenddouble(char **buffer, int *pos,
 	if (zend_isnan(number)) {
 		sign = (number<0);
 		php_sprintf_appendstring(buffer, pos, size, "NaN", 3, 0, padding,
-								 alignment, precision, sign, 0);
+								 alignment, precision, sign, 0, always_sign);
 		return;
 	}
 
 	if (zend_isinf(number)) {
 		sign = (number<0);
 		php_sprintf_appendstring(buffer, pos, size, "INF", 3, 0, padding,
-								 alignment, precision, sign, 0);
+								 alignment, precision, sign, 0, always_sign);
 		return;
 	}
 
 	cvt = php_convert_to_decimal(number, precision, &decpt, &sign, (fmt == 'e'));
+	cvt_len = strlen(cvt);
 
 	if (sign) {
 		numbuf[i++] = '-';
+	} else if (always_sign) {
+		numbuf[i++] = '+';
 	}
 
-	if (fmt == 'f') {
+	if (fmt == 'f' || fmt == 'F') {
 		if (decpt <= 0) {
 			numbuf[i++] = '0';
 			if (precision > 0) {
 				int k = precision;
-				numbuf[i++] = '.';
+				numbuf[i++] = fmt == 'F' ? decimal_point : locale_decimal_point;
 				while ((decpt++ < 0) && k--) {
 					numbuf[i++] = '0';
 				}
 			}
 		} else {
-			while (decpt-- > 0)
+			while (decpt-- > 0) {
+				numbuf[i++] = j < cvt_len ? cvt[j++] : '0';
+			}
+			if (precision > 0) {
+				numbuf[i++] = fmt == 'F' ? decimal_point : locale_decimal_point;
+				while (precision-- > 0) {
+					numbuf[i++] = j < cvt_len ? cvt[j++] : '0';
+				}
+			}
+		}
+	} else if (fmt == 'e' || fmt == 'E') {
+		char *exp_p;
+		int dec2;
+		
+		decpt--;
+		
+		numbuf[i++] = cvt[j++];
+		numbuf[i++] = decimal_point;	
+
+		if (precision > 0) {
+			int k = precision;
+				
+			while (k-- && cvt[j]) {
 				numbuf[i++] = cvt[j++];
-			if (precision > 0)
-				numbuf[i++] = '.';
+			}
+		} else {
+			numbuf[i++] = '0';
+		}
+		
+		numbuf[i++] = fmt;
+		exp_p = php_convert_to_decimal(decpt, 0, &dec2, &sign, 0);
+		numbuf[i++] = sign ? '-' : '+';
+		if (*exp_p) { 
+			while (*exp_p) {
+				numbuf[i++] = *(exp_p++);
+			}
+		} else {
+			numbuf[i++] = '0';
 		}
 	} else {
 		numbuf[i++] = cvt[j++];
 		if (precision > 0)
-			numbuf[i++] = '.';
+			numbuf[i++] = decimal_point;
 	}
 
 	while (cvt[j]) {
@@ -328,21 +406,19 @@ php_sprintf_appenddouble(char **buffer, int *pos,
 
 	numbuf[i] = '\0';
 
-	if (precision > 0) {
-		width += (precision + 1);
-	}
 	php_sprintf_appendstring(buffer, pos, size, numbuf, width, 0, padding,
-							 alignment, i, sign, 0);
+							 alignment, i, sign, 0, always_sign);
 }
 
 
 inline static void
-php_sprintf_append2n(char **buffer, int *pos, int *size, int number,
+php_sprintf_append2n(char **buffer, int *pos, int *size, long number,
 					 int width, char padding, int alignment, int n,
 					 char *chartable, int expprec)
 {
 	char numbuf[NUM_BUF_SIZE];
-	register unsigned int num, i = NUM_BUF_SIZE - 1, neg = 0;
+	register unsigned long num;
+	register unsigned int  i = NUM_BUF_SIZE - 1;
 	register int andbits = (1 << n) - 1;
 
 	PRINTF_DEBUG(("sprintf: append2n(%x, %x, %x, %d, %d, '%c', %d, %d, %x)\n",
@@ -350,13 +426,7 @@ php_sprintf_append2n(char **buffer, int *pos, int *size, int number,
 				  chartable));
 	PRINTF_DEBUG(("sprintf: append2n 2^%d andbits=%x\n", n, andbits));
 
-	if (number < 0) {
-		neg = 1;
-		num = ((unsigned int) -(number + 1)) + 1;
-	} else {
-		num = (unsigned int) number;
-	}
-
+	num = (unsigned long) number;
 	numbuf[i] = '\0';
 
 	do {
@@ -365,20 +435,17 @@ php_sprintf_append2n(char **buffer, int *pos, int *size, int number,
 	}
 	while (num > 0);
 
-	if (neg) {
-		numbuf[--i] = '-';
-	}
 	php_sprintf_appendstring(buffer, pos, size, &numbuf[i], width, 0,
 							 padding, alignment, (NUM_BUF_SIZE - 1) - i,
-							 neg, expprec);
+							 0, expprec, 0);
 }
 
 
-inline static int
+inline static long
 php_sprintf_getnumber(char *buffer, int *pos)
 {
 	char *endptr;
-	register int num = strtol(&buffer[*pos], &endptr, 10);
+	register long num = strtol(&buffer[*pos], &endptr, 10);
 	register int i = 0;
 
 	if (endptr != NULL) {
@@ -389,8 +456,7 @@ php_sprintf_getnumber(char *buffer, int *pos)
 	return num;
 }
 
-
-/*
+/* {{{ php_formatted_print
  * New sprintf implementation for PHP.
  *
  * Modifiers:
@@ -399,6 +465,7 @@ php_sprintf_getnumber(char *buffer, int *pos)
  *  "-"   left adjusted field
  *   n    field size
  *  "."n  precision (floats only)
+ *  "+"   Always place a sign (+ or -) in front of a number
  *
  * Type specifiers:
  *
@@ -406,7 +473,8 @@ php_sprintf_getnumber(char *buffer, int *pos)
  *  "b"   integer argument is printed as binary
  *  "c"   integer argument is printed as a single character
  *  "d"   argument is an integer
- *  "f"   the argument is a float
+ *  "f"   the argument is a float, the decimal separator is locale-aware
+ *  "F"   the argument is a float, but the decimal separator is always "."
  *  "o"   integer argument is printed as octal
  *  "s"   argument is a string
  *  "x"   integer argument is printed as lowercase hexadecimal
@@ -414,51 +482,71 @@ php_sprintf_getnumber(char *buffer, int *pos)
  *
  */
 static char *
-php_formatted_print(int ht, int *len)
+php_formatted_print(int ht, int *len, int use_array TSRMLS_DC)
 {
-	pval ***args;
+	zval ***args, **z_format, **array;
 	int argc, size = 240, inpos = 0, outpos = 0, temppos;
 	int alignment, width, precision, currarg, adjusting, argnum;
 	char *format, *result, padding;
+	int always_sign;
 
 	argc = ZEND_NUM_ARGS();
 
-	if (argc < 1) {
-		WRONG_PARAM_COUNT_WITH_RETVAL(NULL);
-	}
-	args = (pval ***)emalloc(argc * sizeof(pval *));
+	if (use_array) {
+		int i = 1;
 
-	if (zend_get_parameters_array_ex(argc, args) == FAILURE) {
-		efree(args);
-		WRONG_PARAM_COUNT_WITH_RETVAL(NULL);
+		if (ZEND_NUM_ARGS() != 2 || zend_get_parameters_ex(argc, &z_format, &array) == FAILURE) {
+			WRONG_PARAM_COUNT_WITH_RETVAL(NULL);
+		}
+		SEPARATE_ZVAL(array);
+		convert_to_array_ex(array);
+		argc = 1 + zend_hash_num_elements(Z_ARRVAL_PP(array));
+		args = (zval ***)safe_emalloc(argc, sizeof(zval *), 0);
+		args[0] = z_format;
+		for (zend_hash_internal_pointer_reset(Z_ARRVAL_PP(array));
+			 zend_hash_get_current_data(Z_ARRVAL_PP(array), (void **)&args[i++]) == SUCCESS;
+			 zend_hash_move_forward(Z_ARRVAL_PP(array)));
+	} else {
+		if (argc < 1) {
+			WRONG_PARAM_COUNT_WITH_RETVAL(NULL);
+		}
+
+		args = (zval ***)safe_emalloc(argc, sizeof(zval *), 0);
+
+		if (zend_get_parameters_array_ex(argc, args) == FAILURE) {
+			efree(args);
+			WRONG_PARAM_COUNT_WITH_RETVAL(NULL);
+		}
 	}
 	convert_to_string_ex(args[0]);
-	format = (*args[0])->value.str.val;
+	format = Z_STRVAL_PP(args[0]);
 	result = emalloc(size);
 
 	currarg = 1;
 
-	while (inpos<(*args[0])->value.str.len) {
-		int expprec = 0;
+	while (inpos<Z_STRLEN_PP(args[0])) {
+		int expprec = 0, multiuse = 0;
+		zval *tmp;
 
 		PRINTF_DEBUG(("sprintf: format[%d]='%c'\n", inpos, format[inpos]));
 		PRINTF_DEBUG(("sprintf: outpos=%d\n", outpos));
 		if (format[inpos] != '%') {
-			php_sprintf_appendchar(&result, &outpos, &size, format[inpos++]);
+			php_sprintf_appendchar(&result, &outpos, &size, format[inpos++] TSRMLS_CC);
 		} else if (format[inpos + 1] == '%') {
-			php_sprintf_appendchar(&result, &outpos, &size, '%');
+			php_sprintf_appendchar(&result, &outpos, &size, '%' TSRMLS_CC);
 			inpos += 2;
 		} else {
 			if (currarg >= argc && format[inpos + 1] != '%') {
 				efree(result);
 				efree(args);
-				php_error(E_WARNING, "%s(): too few arguments",get_active_function_name());
+				php_error_docref(NULL TSRMLS_CC, E_WARNING, "Too few arguments");
 				return NULL;
 			}
 			/* starting a new format specifier, reset variables */
 			alignment = ALIGN_RIGHT;
 			adjusting = 0;
 			padding = ' ';
+			always_sign = 0;
 			inpos++;			/* skip the '%' */
 
 			PRINTF_DEBUG(("sprintf: first looking at '%c', inpos=%d\n",
@@ -469,6 +557,15 @@ php_formatted_print(int ht, int *len)
 				while (isdigit((int)format[temppos])) temppos++;
 				if (format[temppos] == '$') {
 					argnum = php_sprintf_getnumber(format, &inpos);
+
+					if (argnum == 0) {
+						efree(result);
+						efree(args);
+						php_error_docref(NULL TSRMLS_CC, E_WARNING, "Zero is not a valid argument number");
+						return NULL;
+					}
+
+					multiuse = 1;
 					inpos++;  /* skip the '$' */
 				} else {
 					argnum = currarg++;
@@ -476,7 +573,7 @@ php_formatted_print(int ht, int *len)
 				if (argnum >= argc) {
 					efree(result);
 					efree(args);
-					php_error(E_WARNING, "%s(): too few arguments",get_active_function_name());
+					php_error_docref(NULL TSRMLS_CC, E_WARNING, "Too few arguments");
 					return NULL;
 				}
 
@@ -490,6 +587,8 @@ php_formatted_print(int ht, int *len)
 					} else if (format[inpos] == '-') {
 						alignment = ALIGN_LEFT;
 						/* space padding, the default */
+					} else if (format[inpos] == '+') {
+						always_sign = 1;
 					} else if (format[inpos] == '\'') {
 						padding = format[++inpos];
 					} else {
@@ -537,87 +636,103 @@ php_formatted_print(int ht, int *len)
 			}
 			PRINTF_DEBUG(("sprintf: format character='%c'\n", format[inpos]));
 			/* now we expect to find a type specifier */
+ 			if (multiuse) {
+ 				MAKE_STD_ZVAL(tmp);
+ 				*tmp = **(args[argnum]);
+				INIT_PZVAL(tmp);
+ 				zval_copy_ctor(tmp);
+ 			} else {
+ 				SEPARATE_ZVAL(args[argnum]);
+ 				tmp = *(args[argnum]);
+ 			}
+
 			switch (format[inpos]) {
 				case 's':
-					convert_to_string_ex(args[argnum]);
+					convert_to_string(tmp);
 					php_sprintf_appendstring(&result, &outpos, &size,
-											 (*args[argnum])->value.str.val,
+											 Z_STRVAL_P(tmp),
 											 width, precision, padding,
 											 alignment,
-											 (*args[argnum])->value.str.len,
-											 0, expprec);
+											 Z_STRLEN_P(tmp),
+											 0, expprec, 0);
 					break;
 
 				case 'd':
-					convert_to_long_ex(args[argnum]);
+					convert_to_long(tmp);
 					php_sprintf_appendint(&result, &outpos, &size,
-										  (*args[argnum])->value.lval,
-										  width, padding, alignment);
+										  Z_LVAL_P(tmp),
+										  width, padding, alignment,
+										  always_sign);
 					break;
 
 				case 'u':
-					convert_to_long_ex(args[argnum]);
+					convert_to_long(tmp);
 					php_sprintf_appenduint(&result, &outpos, &size,
-										  (*args[argnum])->value.lval,
+										  Z_LVAL_P(tmp),
 										  width, padding, alignment);
 					break;
 
 				case 'e':
 				case 'f':
+				case 'F':
 					/* XXX not done */
-					convert_to_double_ex(args[argnum]);
+					convert_to_double(tmp);
 					php_sprintf_appenddouble(&result, &outpos, &size,
-											 (*args[argnum])->value.dval,
+											 Z_DVAL_P(tmp),
 											 width, padding, alignment,
 											 precision, adjusting,
-											 format[inpos]);
+											 format[inpos], always_sign
+											 TSRMLS_CC);
 					break;
-
+					
 				case 'c':
-					convert_to_long_ex(args[argnum]);
+					convert_to_long(tmp);
 					php_sprintf_appendchar(&result, &outpos, &size,
-										(char) (*args[argnum])->value.lval);
+										(char) Z_LVAL_P(tmp) TSRMLS_CC);
 					break;
 
 				case 'o':
-					convert_to_long_ex(args[argnum]);
+					convert_to_long(tmp);
 					php_sprintf_append2n(&result, &outpos, &size,
-										 (*args[argnum])->value.lval,
+										 Z_LVAL_P(tmp),
 										 width, padding, alignment, 3,
 										 hexchars, expprec);
 					break;
 
 				case 'x':
-					convert_to_long_ex(args[argnum]);
+					convert_to_long(tmp);
 					php_sprintf_append2n(&result, &outpos, &size,
-										 (*args[argnum])->value.lval,
+										 Z_LVAL_P(tmp),
 										 width, padding, alignment, 4,
 										 hexchars, expprec);
 					break;
 
 				case 'X':
-					convert_to_long_ex(args[argnum]);
+					convert_to_long(tmp);
 					php_sprintf_append2n(&result, &outpos, &size,
-										 (*args[argnum])->value.lval,
+										 Z_LVAL_P(tmp),
 										 width, padding, alignment, 4,
 										 HEXCHARS, expprec);
 					break;
 
 				case 'b':
-					convert_to_long_ex(args[argnum]);
+					convert_to_long(tmp);
 					php_sprintf_append2n(&result, &outpos, &size,
-										 (*args[argnum])->value.lval,
+										 Z_LVAL_P(tmp),
 										 width, padding, alignment, 1,
 										 hexchars, expprec);
 					break;
 
 				case '%':
-					php_sprintf_appendchar(&result, &outpos, &size, '%');
+					php_sprintf_appendchar(&result, &outpos, &size, '%' TSRMLS_CC);
 
 					break;
 				default:
 					break;
 			}
+			if (multiuse) {
+				zval_ptr_dtor(&tmp);
+			}
 			inpos++;
 		}
 	}
@@ -629,6 +744,7 @@ php_formatted_print(int ht, int *len)
 	*len = outpos;	
 	return result;
 }
+/* }}} */
 
 /* {{{ proto string sprintf(string format [, mixed arg1 [, mixed ...]])
    Return a formatted string */
@@ -637,10 +753,25 @@ PHP_FUNCTION(user_sprintf)
 	char *result;
 	int len;
 	
-	if ((result=php_formatted_print(ht,&len))==NULL) {
+	if ((result=php_formatted_print(ht, &len, 0 TSRMLS_CC))==NULL) {
 		RETURN_FALSE;
 	}
-	RETVAL_STRINGL(result,len,1);
+	RETVAL_STRINGL(result, len, 1);
+	efree(result);
+}
+/* }}} */
+
+/* {{{ proto string vsprintf(string format, array args)
+   Return a formatted string */
+PHP_FUNCTION(vsprintf)
+{
+	char *result;
+	int len;
+	
+	if ((result=php_formatted_print(ht, &len, 1 TSRMLS_CC))==NULL) {
+		RETURN_FALSE;
+	}
+	RETVAL_STRINGL(result, len, 1);
 	efree(result);
 }
 /* }}} */
@@ -652,11 +783,28 @@ PHP_FUNCTION(user_printf)
 	char *result;
 	int len;
 	
-	if ((result=php_formatted_print(ht,&len))==NULL) {
+	if ((result=php_formatted_print(ht, &len, 0 TSRMLS_CC))==NULL) {
 		RETURN_FALSE;
 	}
-	PHPWRITE(result,len);
+	PHPWRITE(result, len);
 	efree(result);
+	RETURN_LONG(len);
+}
+/* }}} */
+
+/* {{{ proto int vprintf(string format, array args)
+   Output a formatted string */
+PHP_FUNCTION(vprintf)
+{
+	char *result;
+	int len;
+	
+	if ((result=php_formatted_print(ht, &len, 1 TSRMLS_CC))==NULL) {
+		RETURN_FALSE;
+	}
+	PHPWRITE(result, len);
+	efree(result);
+	RETURN_LONG(len);
 }
 /* }}} */
 
@@ -665,4 +813,6 @@ PHP_FUNCTION(user_printf)
  * tab-width: 4
  * c-basic-offset: 4
  * End:
+ * vim600: sw=4 ts=4 fdm=marker
+ * vim<600: sw=4 ts=4
  */
